#!/usr/bin/env python
import click
import glob
import time
import os
from os.path import join as join_paths, isfile

from graphql import GraphQLSchema
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler, EVENT_TYPE_CREATED, EVENT_TYPE_MODIFIED

from gql.config import Config
from gql.query_parser import QueryParser, AnonymousQueryError, InvalidQueryError
from gql.renderer_dataclasses import DataclassesRenderer
from gql.utils_schema import load_schema

DEFAULT_CONFIG_FNAME = '.gql.json'
SCHEMA_PROMPT = click.style('Where is your schema?: ', fg='bright_white') + \
                click.style('(path or url) ', fg='bright_black', dim=False)

ROOT_PROMPT = click.style('Whats the root of your project: ', fg='bright_white') + \
              click.style('(path or url) ', fg='bright_black', dim=False)


def safe_remove(fname):
    try:
        os.remove(fname)
    except:
        pass


@click.group()
def cli():
    pass


@cli.command()
@click.option('--schema', prompt=SCHEMA_PROMPT, default='http://localhost:4000')
@click.option('--endpoint', prompt=SCHEMA_PROMPT, default='same as schema')
@click.option('--root', prompt=ROOT_PROMPT, default='./src')
@click.option('-c', '--config', 'config_filename', default=DEFAULT_CONFIG_FNAME, type=click.Path(exists=False))
def init(schema, endpoint, root, config_filename):
    if isfile(config_filename):
        click.confirm(f'{config_filename} already exists. Are you sure you want to continue?', abort=True)

    if endpoint == 'same as schema':
        endpoint = schema

    config = Config(
        schema=schema,
        endpoint=endpoint,
        documents=join_paths(root, '**/*.graphql')
    )

    config.save(config_filename)

    click.echo(f"Config file generated at {click.style(config_filename, fg='bright_white')}\n\n")


def process_file(filename: str, parser: QueryParser, renderer: DataclassesRenderer):
    root, _s = os.path.splitext(filename)
    target_filename = root + '.py'

    click.echo(f'Parsing {filename} ... ', nl=False)
    with open(filename, 'r') as fin:
        query = fin.read()
        try:
            parsed = parser.parse(query)
            rendered = renderer.render(parsed)
            with open(target_filename, 'w') as outfile:
                outfile.write(rendered)
                click.secho('Success!', fg='bright_white')

        except AnonymousQueryError:
            click.secho('Failed!', fg='bright_red')
            click.secho('\tQuery is missing a name', fg='bright_black')
            safe_remove(target_filename)
        except InvalidQueryError as invalid_err:
            click.secho('Failed!', fg='bright_red')
            click.secho(f'\t{invalid_err}', fg='bright_black')
            safe_remove(target_filename)


@cli.command()
@click.option('-c', '--config', 'config_filename', default=DEFAULT_CONFIG_FNAME, type=click.Path(exists=True))
def run(config_filename):
    if not isfile(config_filename):
        click.echo(f'Could not find configuration file {config_filename}')

    config = Config.load(config_filename)
    schema = load_schema(config.schema)

    filenames = glob.glob(config.documents, recursive=True)

    query_parser = QueryParser(schema)
    query_renderer = DataclassesRenderer(schema, config)

    for filename in filenames:
        process_file(filename, query_parser, query_renderer)


@cli.command()
@click.option('-c', '--config', 'config_filename', default=DEFAULT_CONFIG_FNAME, type=click.Path(exists=True))
def watch(config_filename):
    class Handler(FileSystemEventHandler):
        def __init__(self, config: Config, schema: GraphQLSchema):
            self.parser = QueryParser(schema)
            self.renderer = DataclassesRenderer(schema, config)

        def on_any_event(self, event):
            if event.is_directory:
                return

            if event.event_type in {EVENT_TYPE_CREATED, EVENT_TYPE_MODIFIED}:
                filenames = [os.path.abspath(fn) for fn in glob.iglob(config.documents, recursive=True)]
                if event.src_path not in filenames:
                    return

                # Take any action here when a file is first created.
                process_file(event.src_path, self.parser, self.renderer)

    if not isfile(config_filename):
        click.echo(f'Could not find configuration file {config_filename}')

    config = Config.load(config_filename)
    schema = load_schema(config.schema)

    click.secho(f'Watching {config.documents}', fg='cyan')
    click.secho('Ready for changes...', fg='cyan')

    observer = Observer()
    observer.schedule(Handler(config, schema), os.path.abspath('./'), recursive=True)
    observer.start()
    try:
        while True:
            time.sleep(5)
    except:
        observer.stop()
        print('Error')

    observer.join()


if __name__ == '__main__':
    cli()



from graphql import GraphQLSchema

from gql.config import Config
from gql.utils_codegen import CodeChunk
from gql.query_parser import ParsedQuery, ParsedField, ParsedObject, ParsedEnum, ParsedOperation, ParsedVariableDefinition


class DataclassesRenderer:

    def __init__(self, schema: GraphQLSchema, config: Config):
        self.schema = schema
        self.config = config

    def render(self, parsed_query: ParsedQuery):
        # We sort fragment nodes to be first and operations to be last because of dependecies
        buffer = CodeChunk()
        buffer.write('# AUTOGENERATED file. Do not Change!')
        buffer.write('from functools import partial')
        buffer.write('from typing import Any, Callable, Mapping, List')
        buffer.write('from enum import Enum')
        buffer.write('from dataclasses import dataclass, field')
        buffer.write('from dataclasses_json import dataclass_json')
        buffer.write('from gql.clients import Client, AsyncIOClient')
        buffer.write('')

        if self.config.custom_header:
            buffer.write_lines(self.config.custom_header.split('\n'))

        buffer.write('')

        self.__render_datetime_field(buffer)

        # Enums
        if parsed_query.enums:
            self.__render_enum_field(buffer)
            for enum in parsed_query.enums:
                self.__render_enum(buffer, enum)

        sorted_objects = sorted(parsed_query.objects, key=lambda obj: 1 if isinstance(obj, ParsedOperation) else 0)
        for obj in sorted_objects:
            if isinstance(obj, ParsedObject):
                self.__render_object(parsed_query, buffer, obj)
            elif isinstance(obj, ParsedOperation):
                self.__render_operation(parsed_query, buffer, obj)

        return str(buffer)

    @staticmethod
    def __render_enum_field(buffer: CodeChunk):
        with buffer.write_block('def enum_field(enum_type):'):
            with buffer.write_block('def encode_enum(value):'):
                buffer.write('return value.value')
            buffer.write('')
            with buffer.write_block('def decode_enum(t, value):'):
                buffer.write('return t(value)')

            buffer.write('')
            buffer.write("return field(metadata={'dataclasses_json': {'encoder': encode_enum, 'decoder': partial(decode_enum, enum_type)}})")
            buffer.write('')

    @staticmethod
    def __render_datetime_field(buffer: CodeChunk):
        buffer.write('')
        buffer.write('from datetime import datetime')
        buffer.write('from marshmallow import fields as marshmallow_fields')
        buffer.write("DATETIME_FIELD = field(metadata={'dataclasses_json': {'encoder': datetime.isoformat, 'decoder': datetime.fromisoformat, 'mm_field': marshmallow_fields.DateTime(format='iso')}})")
        buffer.write('')

    def __render_object(self, parsed_query: ParsedQuery, buffer: CodeChunk, obj: ParsedObject):
        class_parents = '' if not obj.parents else f'({", ".join(obj.parents)})'

        buffer.write('@dataclass_json')
        buffer.write('@dataclass')
        with buffer.write_block(f'class {obj.name}{class_parents}:'):
            # render child objects
            for child_object in obj.children:
                self.__render_object(parsed_query, buffer, child_object)

            # render fields
            sorted_fields = sorted(obj.fields, key=lambda f: 1 if f.nullable else 0)
            for field in sorted_fields:
                self.__render_field(parsed_query, buffer, field)

            # pass if not children or fields
            if not (obj.children or obj.fields):
                buffer.write('pass')

        buffer.write('')

    def __render_operation(self, parsed_query: ParsedQuery, buffer: CodeChunk, parsed_op: ParsedOperation):
        buffer.write('@dataclass_json')
        buffer.write('@dataclass')
        with buffer.write_block(f'class {parsed_op.name}:'):
            buffer.write('__QUERY__ = """')
            buffer.write(parsed_query.query)
            buffer.write('"""')
            buffer.write('')

            # Render children
            for child_object in parsed_op.children:
                self.__render_object(parsed_query, buffer, child_object)

            # operation fields
            buffer.write('')
            buffer.write(f'data: {parsed_op.name}Data = None')
            buffer.write('errors: Any = None')
            buffer.write('')

            # Execution functions
            if parsed_op.variables:
                vars_args = ', '.join([self.__render_variable_definition(var) for var in parsed_op.variables]) + ','
                variables_dict = '{' + ', '.join(f'"{var.name}": {var.name}' for var in parsed_op.variables) + '}'
            else:
                vars_args = ''
                variables_dict = 'None'

            buffer.write('@classmethod')
            with buffer.write_block(f'def execute(cls, {vars_args} on_before_callback: Callable[[Mapping[str, str], Mapping[str, str]], None] = None):'):
                buffer.write(f'client = Client(\'{self.config.endpoint}\')')
                buffer.write(f'variables = {variables_dict}')
                buffer.write('response_text = client.call(cls.__QUERY__, variables=variables, on_before_callback=on_before_callback)')
                buffer.write('return cls.from_json(response_text)')

            buffer.write('')

            buffer.write('@classmethod')
            with buffer.write_block(f'async def execute_async(cls, {vars_args} on_before_callback: Callable[[Mapping[str, str], Mapping[str, str]], None] = None):'):
                buffer.write(f'client = AsyncIOClient(\'{self.config.endpoint}\')')
                buffer.write(f'variables = {variables_dict}')
                buffer.write(f'response_text = await client.call(cls.__QUERY__, variables=variables, on_before_callback=on_before_callback)')
                buffer.write(f'return cls.from_json(response_text)')

            buffer.write('')
            buffer.write('')

    @staticmethod
    def __render_variable_definition(var: ParsedVariableDefinition):
        if not var.nullable:
            return f'{var.name}: {var.type}'

        return f'{var.name}: {var.type} = {var.default_value or "None"}'

    @staticmethod
    def __render_field(parsed_query: ParsedQuery, buffer: CodeChunk, field: ParsedField):
        enum_names = [e.name for e in parsed_query.enums]
        is_enum = field.type in enum_names
        suffix = ''
        field_type = field.type

        if is_enum:
            suffix = f'= enum_field({field.type})'

        if field.type == 'DateTime':
            suffix = '= DATETIME_FIELD'
            field_type = 'datetime'

        if field.nullable:
            suffix = f'= {field.default_value}'

        buffer.write(f'{field.name}: {field_type} {suffix}')

    @staticmethod
    def __render_enum(buffer: CodeChunk, enum: ParsedEnum):
        with buffer.write_block(f'class {enum.name}(Enum):'):
            for value_name, value in enum.values.items():
                if isinstance(value, str):
                    value = f"'{value}'"

                buffer.write(f'{value_name} = {value}')

        buffer.write('')



# -*- coding: utf-8 -*-
from distutils.core import setup

package_dir = \
{'': '.'}

packages = \
['gql', 'gql.clients']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0',
 'dataclasses-json>=0.2.0,<0.3.0',
 'graphql-core-next>=1.0.0,<1.1.0',
 'inflection>=0.3.1,<0.4.0',
 'jinja2>=2.10,<3.0',
 'requests>=2.21,<3.0',
 'watchdog>=0.9.0,<0.10.0']

extras_require = \
{'async': ['aiohttp>=3.5,<4.0']}

entry_points = \
{'console_scripts': ['gql = gql.cli:cli']}

setup_kwargs = {
    'name': 'gql-next',
    'version': '0.1.1',
    'description': 'Python GraphQL Client Library',
    'long_description': '# GQL: Python GraphQL Client Library\n\n[![Build Status](https://travis-ci.org/ekampf/cql.svg?branch=master)](https://travis-ci.org/ekampf/cql)\n[![Coverage Status](https://coveralls.io/repos/github/ekampf/gql/badge.svg?branch=master)](https://coveralls.io/github/ekampf/gql?branch=master)\n\n## Introduction\n\nGQL is a GraphQL Client Python library intended to help Python application make GraphQL\nAPI call while enjoying the advantages that come with GraphQL.\n\n- **Strongly Typed** response objects (dynamically created in build time to match your query)\n- **Query Validation** that checks your code\'s queries against the GraphQL server\'s schema.\n\n## Installation\n\nSimply install from PyPi:\n\n```bash\npip install gql-next\n```\n\nThen go to your project folder and run `gql init`\n\n## Quick Start\n\n`gql` works by parsing query files (`**/*.graphql` by default) into their own Python module where\nan class, named after the operation defined in the file, allows you to make that query and get a typed\nresponse.\n\nFor example, given the following file `get_film.graphql` file:\n```\nquery GetFilm($id: ID!) {\n  film(id: $id) {\n    title\n    director\n  }\n}\n```\n\nA `get_film.py` will be created defining a `GetFilm` class:\n\n```python\n# AUTOGENERATED file. Do not Change!\nfrom typing import Any, Callable, Mapping, List\nfrom enum import Enum\nfrom dataclasses import dataclass\nfrom dataclasses_json import dataclass_json\nfrom gql.clients import Client, AsyncIOClient\n\n\n@dataclass_json\n@dataclass\nclass GetFilm:\n    @dataclass_json\n    @dataclass\n    class GetFilmData:\n        @dataclass_json\n        @dataclass\n        class Film:\n            title: str\n            director: str\n        film: Film = None\n\n    data: GetFilmData = None\n    errors: Any = None\n\n    @classmethod\n    def execute(cls, id: str, on_before_callback: Callable[[Mapping[str, str], Mapping[str, str]], None] = None) -> GetFilm:\n        ...\n\n    @classmethod\n    async def execute_async(cls, id: str, on_before_callback: Callable[[Mapping[str, str], Mapping[str, str]], None] = None) -> GetFilm:\n        ...\n```\n\nAllowing you to make the GraphQL query:\n\n```python\nfrom .get_film import GetFilm\n\nresult = GetFilm.execute(\'meaning_of_life\')\nfilm = result.data.film\n```\n\n*Important notes:*\n* Operations defined in graphql query __must be named__ so that we can name the relevant Python Class which you can then import in your code\n\n\n## How it works\n\n\n### The `gql` client\n\n#### `gql init`\nInitializes a project to use GQL as client - writes a .gql.json configuration file.\n\n#### `gql run`\n\nRun through your project\'s files and compile GraphQL queries into into Python types.\n\n#### `gql watch`\n\nUseful during development. Listen to file changes in your project\'s folder and continuously\nbuilds GraphQL queries as they change.\nThis allows you to:\n* Immediately verify query changes you make are valid.\n* Enjoy your IDE\'s autocomplete features on GraphQL auto-generated objects while developing\nas `watch` will auto-update them as you change queries.\n\n\n# Sponsors\n\n<a href="https://ebates.com"><img src="https://opensource.ebates.com/static/images/ebates-rakuten.svg" width="250"></a>\n',
    'author': 'Eran Kampf',
    'author_email': 'eran@ekampf.com',
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
